use pest_derive::Parser;

#[derive(Parser)]
#[grammar = "parser/deimos.pest"]
pub struct DeimosParser;

#[cfg(test)]
mod deimos_parser_tests {
    use crate::parser::{DeimosParser, Rule};
    use indoc::indoc;
    use pest::Parser;

    macro_rules! fail_rule {
        ($pair: expr; $rule:path) => {{
            panic!(
                "Expected {} but found {:?}",
                stringify!($rule),
                $pair.as_rule()
            )
        }};
    }

    macro_rules! match_rule {
        ($pair:expr; $rule:path) => {{
            if $rule != $pair.as_rule() {
                fail_rule!($pair; $rule);
            }
        }};
    }

    fn parses_to(rule: Rule, input: &str) {
        let parse_result = DeimosParser::parse(rule, input);
        if let Err(e) = parse_result {
            panic!("Parsing failed.\n{}", e);
        } else {
            let mut pairs = parse_result.unwrap();
            let first = pairs.next().unwrap();
            match_rule!(first; rule)
        }
    }

    #[test]
    fn hex_int() {
        parses_to(Rule::IntLiteral, "0x1234abcd");
    }

    #[test]
    fn hex_long() {
        parses_to(Rule::LongLiteral, "0x123456789abcdefL");
    }

    #[test]
    fn suffix_expression_call_single_identifier() {
        parses_to(Rule::SuffixExpression, "foo()");
    }

    #[test]
    fn simple_interface() {
        parses_to(Rule::CompilationUnit, "pub int Simple { fn foo() -> Void }");
    }

    #[test]
    fn interface_with_op() {
        parses_to(
            Rule::CompilationUnit,
            "pub int Callable { op () () -> Void }",
        );
    }

    #[test]
    fn interface_with_alias() {
        parses_to(
            Rule::CompilationUnit,
            indoc! {"
                pub int Callable {
                    fn call() -> Void
                    def op () () -> Void alias call
                }
            "},
        );
    }

    #[test]
    fn index_identifier() {
        parses_to(Rule::SuffixExpression, "foo[0]");
    }

    #[test]
    fn chained_index_call_on_identifier() {
        parses_to(Rule::SuffixExpression, "foo[0]()");
    }
}
